@@ -1,8 +1,10 @@
-import argparse
 from dataclasses import dataclass
 from enum import Enum
 from typing import Any
 
+from fastled.args import Args
+from fastled.print_filter import PrintFilterDefault
+
 
 @dataclass
 class CompileResult:
@@ -17,6 +19,11 @@ class CompileResult:
     def to_dict(self) -> dict[str, Any]:
         return self.__dict__.copy()
 
+    def __post_init__(self):
+        # Filter the stdout.
+        pf = PrintFilterDefault(echo=False)
+        self.stdout = pf.print(self.stdout)
+
 
 class CompileServerError(Exception):
     """Error class for failing to instantiate CompileServer."""
@@ -38,7 +45,7 @@ class BuildMode(Enum):
             raise ValueError(f"BUILD_MODE must be one of {valid_modes}, got {mode_str}")
 
     @staticmethod
-    def from_args(args: argparse.Namespace) -> "BuildMode":
+    def from_args(args: Args) -> "BuildMode":
         if args.debug:
             return BuildMode.DEBUG
         elif args.release:
@@ -59,3 +66,12 @@ class Platform(Enum):
             raise ValueError(
                 f"Platform must be one of {valid_modes}, got {platform_str}"
             )
+
+
+@dataclass
+class FileResponse:
+    """File response from the server."""
+
+    filename: str
+    content: str
+    mimetype: str
